#include <stdio.h>
#include <windows.h>
#include <iostream>

using namespace std;

//#define M 110
//#define k 12


#define M 100
#define k 12

//#define M 240
//#define k 4

void showMatrix(double *matrix, int matrixSize, int width);

double* MatrixMultiplication(double*  matrix_A, double*  matrix_B);
double* MatrixMultiplication_asm(double*  matrix_A, double*  matrix_B);
double* MatrixMultiplication_intrinsic(double*  matrix_A, double*  matrix_B);


int main()
{
	size_t matrixByteSize = M * M * k * k * sizeof(double);
	double*  matrix_A = (double*)_aligned_malloc(matrixByteSize, 16);
	double*  matrix_B = (double*)_aligned_malloc(matrixByteSize, 16);

	if(matrix_A == NULL || matrix_B == NULL)
	{
		cout << "Allocate error!\n";
		system("pause");
		return 0;
	}

	int matrixSize = M * M * k * k;

	//  
	for(int i = 0; i < matrixSize; i++)
	{
		matrix_A[i] = i + 1;
		matrix_B[i] = i + 1;
	}

	double* result_vectorized = MatrixMultiplication(matrix_A, matrix_B);
	double* result_asm = MatrixMultiplication_intrinsic(matrix_A, matrix_B);

	
	bool isSame = true;
	
	for(int i = 0; i < matrixSize; i++)
	{
		if(result_vectorized[i] != result_asm[i])
		{
			isSame = false;
			break;
		}
	}
	
	if(isSame)
		cout << "The results are the same\n";
	else
		cout << "The results do not match\n";

	
//	showMatrix(matrix_A, matrixSize, M * k);
//	showMatrix(matrix_B, matrixSize, M * k);
//	showMatrix(result_asm, matrixSize, M * k);
	

	system("pause");
	return 0;
}



double* MatrixMultiplication(double*  matrix_A, double*  matrix_B)
{
	size_t matrixByteSize = M * M * k * k * sizeof(double);
	double*  matrix_C = (double*)_aligned_malloc(matrixByteSize, 16);

	int matrixSize = M * M * k * k;

	//   
	for(int i = 0; i < matrixSize; i++)
		matrix_C[i] = 0;

	DWORD startTime = timeGetTime();

	//   A  B,     C
	for(int i = 0; i < M; i++)
	{
		for(int j = 0; j < M; j++)
		{
			int index = i * M * k * k + j * k;
			for(int step = 0; step < M; step++)
			{
				int index1 = i * M * k * k + step * k;
				int index2 = step * M * k * k + j * k;

				for(int i_in = 0; i_in < k; i_in++)
				{			
					for(int step_in = 0; step_in < k; step_in++)
					{						
						//double tmp = matrix_A[index1 + i_in * (M + k)+ step_in];
						//#pragma vector always
						//#pragma simd vectorlengthfor(double)
						for(int j_in = 0; j_in < k; j_in++)
						{
							matrix_C[index + i_in * (M * k) + j_in] += 
								matrix_A[index1 + i_in * (M * k)+ step_in] * matrix_B[index2 + step_in * (M * k) + j_in];
						}
					}
				}
			}
		}
	}

	DWORD executionTime = timeGetTime() - startTime;
	cout << "Execution time of vectorized loops is " << executionTime << " ms\n\n";

	return matrix_C;
}


double* MatrixMultiplication_intrinsic(double*  matrix_A, double*  matrix_B)
{
	size_t matrixByteSize = M * M * k * k * sizeof(double);
	double*  matrix_C = (double*)_aligned_malloc(matrixByteSize, 16);

	int matrixSize = M * M * k * k;

	//   
	for(int i = 0; i < matrixSize; i++)
		matrix_C[i] = 0;

	DWORD startTime = timeGetTime();


	//   A  B,     C
	int i, j, index, index1, index2, t1, i_in, j_in, step_in, step;
	
	for(i = 0; i < M; i++)
	{
		for(j = 0; j < M; j++)
		{
			index = i * M * k * k + j * k;
			for(step = 0; step < M; step++)
			{
				index1 = i * M * k * k + step * k;
				index2 = step * M * k * k + j * k;

				for(i_in = 0; i_in < k; i_in++)
				{			
					t1 = index1 + i_in * (M * k);
					for(step_in = 0; step_in < k; step_in++)
					{						
						__m128d mA = _mm_set_pd(matrix_A[t1 + step_in], matrix_A[t1 + step_in]);
						
						int tmp = index + i_in * (M * k);
						int tmp2 = index2 + step_in * (M * k);

						for(j_in = 0; j_in < k; j_in += 2)
						{					
							__m128d c = _mm_load_pd(matrix_C + tmp + j_in);
							__m128d b = _mm_load_pd(matrix_B + tmp2 + j_in);

							b = _mm_mul_pd(b, mA);
							c = _mm_add_pd(c, b);

							_mm_store_pd(matrix_C + tmp + j_in, c);
						}
					}
				}		
			}
		}
	}

	DWORD executionTime = timeGetTime() - startTime;
	cout << "Execution time of intrinsic loops is " << executionTime << " ms\n\n";

	return matrix_C;
}

void showMatrix(double *matrix, int matrixSize, int width)
{
	cout << "\n\n";
	for(int i = 0; i < matrixSize; i++)
	{
		cout << "\t" << matrix[i];

		if((i + 1) % width == 0)
			cout << endl;
	}
}