#include "DiscreteLaplaceOperator_CPU.h"

#include "hr_time.h"

//     
const int laplaceConvolutionKernel[3][3] = 
{
	{1,  1,  1},
	{1, -8,  1},
	{1,  1,  1}
};



void cpu_Laplace(_uchar4 *dst, _uchar4 *src, int imageW, int imageH)
{
	int pos;
	for (int y = 0; y < imageH; y++) //  
	{
		for (int x = 0; x < imageW; x++) //   
		{
			int rSum = 0, gSum = 0, bSum = 0;

			for (int j = 0; j < 3; j++)
			{
				for (int i = 0; i < 3; i++)
				{
					int pixelPosX = x + (i - 1);
					int pixelPosY = y + (j - 1);

					if (pixelPosX == imageW) pixelPosX = imageW - 1;
					if (pixelPosX == -1) pixelPosX = 0;
					if (pixelPosY == imageH) pixelPosY = imageH - 1;
					if (pixelPosY == -1) pixelPosY = 0;

					pos = imageW * pixelPosY + pixelPosX;

					rSum += src[pos].x * laplaceConvolutionKernel[j][i];
					gSum += src[pos].y * laplaceConvolutionKernel[j][i];
					bSum += src[pos].z * laplaceConvolutionKernel[j][i];
				}
			}
			if (rSum < 0) rSum *= -1;
			if (rSum > 255) rSum = 255;

			if (gSum < 0) gSum *= -1;
			if (gSum > 255) gSum = 255;

			if (bSum < 0) bSum *= -1;
			if (bSum > 255) bSum = 255;

			//    
			_uchar4 result = {(byte)rSum, (byte)gSum, (byte)bSum, 0};
			dst[imageW * y + x] = result;
		}
	}
}



double cpu_Laplace_withTime(_uchar4 *dst, _uchar4 *src, int imageW, int imageH)
{
	CStopWatch timer;
	timer.startTimer();

	cpu_Laplace(dst, src, imageW, imageH);

	timer.stopTimer();

	return timer.getElapsedTime_milliseconds();
}



/*
//     CPU
BYTE* DiscreteLaplaceOperator_CPU(BYTE* pixelData, int width, int byteWidth, int height)
{
	BYTE* result = new BYTE[byteWidth * height]();
	if(result == NULL)
		return NULL;

	int pos;
	for (int y = 0; y < height; y++) //  
	{
		for (int x = 0; x < width; x++) //   
		{
			int rSum = 0, gSum = 0, bSum = 0;

			for (int j = 0; j < 3; j++)
			{
				for (int i = 0; i < 3; i++)
				{
					int pixelPosX = x + (i - 1);
					int pixelPosY = y + (j - 1);

					if (pixelPosX == width) pixelPosX = width - 1;
					if (pixelPosX == -1) pixelPosX = 0;
					if (pixelPosY == height) pixelPosY = height - 1;
					if (pixelPosY == -1) pixelPosY = 0;

					pos = byteWidth * pixelPosY + pixelPosX * 3;
					int kernelVal = laplaceConvolutionKernel[j][i];

					rSum += pixelData[pos + 0] * kernelVal;
					gSum += pixelData[pos + 1] * kernelVal;
					bSum += pixelData[pos + 2] * kernelVal;

					//kSum += kernelVal;
				}
			}
			if (rSum < 0) rSum *= -1;
			if (rSum > 255) rSum = 255;

			if (gSum < 0) gSum *= -1;
			if (gSum > 255) gSum = 255;

			if (bSum < 0) bSum *= -1;
			if (bSum > 255) bSum = 255;

			//    
			pos = byteWidth * y + x * 3;
			result[pos + 0] = (byte)rSum;
			result[pos + 1] = (byte)gSum;
			result[pos + 2] = (byte)bSum;
		}
	}
	return result;
}*/